from subwiz import run


def test_languages():
    results = run(
        input_domains=["german.hadrian.io", "italian.hadrian.io"],
        num_predictions=100,
        no_resolve=True,
    )
    print(results)
    assert "english.hadrian.io" in results


def test_numbers():
    results = run(
        input_domains=["test1.hadrian.io", "test2.hadrian.io"],
        num_predictions=10,
        no_resolve=True,
    )
    assert "test3.hadrian.io" in results



"""Exposes a command line interface that runs subwiz and returns the result."""

import argparse

from subwiz.main import run
from subwiz.type import (
    Domain,
    device_type,
    input_domains_file_type,
    max_recursion_type,
    output_file_type,
    positive_int_type,
    temperature_type,
)

parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument(
    "-i",
    "--input-file",
    help="file containing new-line-separated subdomains.",
    dest="input_file",
    required=True,
    type=input_domains_file_type,
)
parser.add_argument(
    "-o",
    "--output-file",
    help="output file to write new-line separated subdomains to.",
    dest="output_file",
    type=output_file_type,
)
parser.add_argument(
    "-n",
    "--num-predictions",
    help="number of subdomains to predict.",
    dest="num_predictions",
    default=500,
    type=positive_int_type,
)
parser.add_argument(
    "--no-resolve",
    help="do not resolve the output subdomains.",
    dest="no_resolve",
    action="store_true",
)
parser.add_argument(
    "--force-download",
    help="download model and tokenizer files, even if cached.",
    dest="force_download",
    action="store_true",
)
parser.add_argument(
    "--max-recursion",
    help="maximum number of times the inference process will recursively re-run after finding new subdomains.",
    dest="max_recursion",
    default=5,
    type=max_recursion_type,
)
parser.add_argument(
    "-t",
    "--temperature",
    help="add randomness to the model (recommended ≤ 0.3).",
    dest="temperature",
    default=0.0,
    type=temperature_type,
)
parser.add_argument(
    "-d",
    "--device",
    help="hardware to run the transformer model on.",
    dest="device",
    default="auto",
    choices=["auto", "cpu", "cuda", "mps"],
    type=device_type,
)
parser.add_argument(
    "-m",
    "--max-new-tokens",
    help="maximum length of predicted subdomains in tokens.",
    dest="max_new_tokens",
    default=10,
    type=positive_int_type,
)
parser.add_argument(
    "--resolution-concurrency",
    help="number of concurrent resolutions.",
    dest="resolution_concurrency",
    default=128,
    type=positive_int_type,
)
parser.add_argument(
    "--multi-apex",
    help="allow multiple apex domains in the input file. runs inference for each apex separately.",
    dest="multi_apex",
    action="store_true",
)
parser.add_argument(
    "-q",
    "--quiet",
    help="useful for piping into another tool.",
    action="store_true",
)
parser.add_argument(
    "-s",
    "--silent",
    help="do not print any output. requires --output-file.",
    action="store_true",
)
args = parser.parse_args()


def main():
    try:
        if args.silent and not args.output_file:
            parser.error("--silent requires --output-file.")

        domain_objects: list[Domain] = args.input_file
        input_domains = [str(dom) for dom in domain_objects]

        run_args = {
            k: v
            for k, v in args.__dict__.items()
            if k
            not in {
                "input_file",
                "output_file",
                "profile",
                "profile_output",
                "silent",
                "quiet",
            }
        }

        results = run(
            **run_args,
            input_domains=input_domains,
            quiet=args.quiet or args.silent,
        )

        output = "\n".join(sorted(results))

        if args.output_file:
            with open(args.output_file, "w") as f:
                f.write(output)

        if not args.silent:
            print(output)

    except argparse.ArgumentTypeError as e:
        parser.error(str(e))


if __name__ == "__main__":
    main()



import argparse
import asyncio
import os
from typing import Optional, Union

import aiodns
import idna.core
import tldextract
import torch


class Domain:
    def __init__(self, value: str):
        extract = tldextract.extract(value.lower())

        if not (extract.domain and extract.suffix):
            raise ValueError(f"not a valid domain: {value}")

        self._subdomain, self._domain, self._suffix = (
            extract.subdomain,
            extract.domain,
            extract.suffix,
        )

    @property
    def subdomain(self) -> str:
        return self._subdomain

    @property
    def apex_domain(self) -> str:
        return self._domain + "." + self._suffix

    def __str__(self):
        if not self._subdomain:
            return self.apex_domain
        return self._subdomain + "." + self.apex_domain

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == str(other)

    async def is_registered(
        self, resolver: aiodns.DNSResolver, semaphore: asyncio.Semaphore
    ) -> bool:
        async with semaphore:
            try:
                await resolver.query(str(self), "A")
                return True
            except idna.IDNAError:
                return False
            except aiodns.error.DNSError:
                return False


def max_recursion_type(value: str | int) -> int:
    try:
        ivalue = int(value)
    except ValueError:
        raise argparse.ArgumentTypeError(f"use an integer >= 0 and <= 50: {value}")
    # don't allow > 50, so that model does not generate forever for wild card
    if not 0 <= ivalue <= 50:
        raise argparse.ArgumentTypeError(f"use an integer >= 0 and <= 50 {value}")
    return ivalue


def positive_int_type(value: str | int) -> int:
    try:
        ivalue = int(value)
    except ValueError:
        raise argparse.ArgumentTypeError(f"use a positive integer: {value}")
    if ivalue <= 0:
        raise argparse.ArgumentTypeError(f"use a positive integer: {value}")
    return ivalue


def input_domains_file_type(value: str | os.PathLike) -> list[Domain]:
    if not isinstance(value, (os.PathLike, str)):
        raise argparse.ArgumentTypeError(
            f"use a pathlike input for file path:  {value}"
        )
    if not os.path.exists(value):
        raise argparse.ArgumentTypeError(f"file not found: {value}")

    with open(value) as f:
        input_domains = f.read().split()

    return input_domains_type(input_domains)


def input_domains_type(value: list[str]) -> list[Domain]:
    if len(value) == 0:
        raise argparse.ArgumentTypeError("empty input domains")

    value = set(value)

    domains, invalid_domains = [], []
    for dom in value:
        try:
            domains.append(Domain(dom))
        except ValueError:
            invalid_domains.append(dom)

    if invalid_domains:
        raise argparse.ArgumentTypeError(
            f"invalid input domains: {sorted(invalid_domains)}"
        )

    return domains


def output_file_type(value: Optional[Union[str, os.PathLike]]) -> Optional[os.PathLike]:
    if value is None:
        return value

    if not isinstance(value, (os.PathLike, str)):
        raise argparse.ArgumentTypeError(
            f"use a pathlike input for file path:  {value}"
        )
    return value


def temperature_type(value: str | float) -> float:
    try:
        fvalue = float(value)
    except ValueError:
        raise argparse.ArgumentTypeError(f"not a valid float: {value}")
    if not 0 <= fvalue <= 1:
        raise argparse.ArgumentTypeError(f"use 0 ≤ temperature ≤ 1: {value}")
    return fvalue


def concurrency_type(value: str | int) -> int:
    try:
        ivalue = int(value)
    except ValueError:
        raise argparse.ArgumentTypeError(f"not a valid int: {value}")
    if not 1 <= ivalue <= 256:
        raise argparse.ArgumentTypeError(f"use 1 ≤ concurrency ≤ 256: {value}")
    return ivalue


def device_type(value: str) -> str:
    if not isinstance(value, str):
        raise argparse.ArgumentTypeError("use a string for the device")

    if value not in ["auto", "cpu", "cuda", "mps"]:
        raise argparse.ArgumentTypeError(
            f'device should be in ["auto", "cpu", "cuda", "mps"]: {value}'
        )

    if value != "auto":
        return value

    elif torch.cuda.is_available():
        return "cuda"
    elif torch.backends.mps.is_available():
        return "mps"
    return "cpu"



from argparse import ArgumentTypeError

from subwiz import run


def test_zero():
    args = {
        "input_domains": ["admin.hadrian.io", "test.hadrian.io"],
        "num_predictions": 10,
        "temperature": 0,
    }
    first_results = run(**args, no_resolve=True)
    second_results = run(**args, no_resolve=True)

    assert first_results == second_results


def test_high():
    args = {
        "input_domains": ["admin.hadrian.io", "test.hadrian.io"],
        "num_predictions": 20,
        "temperature": 1,
    }
    first_results = run(**args, no_resolve=True)
    second_results = run(**args, no_resolve=True)

    assert first_results != second_results


def test_out_of_bounds():
    try:
        run(
            input_domains=["admin.hadrian.io"],
            temperature=1.1,
            num_predictions=1,
            no_resolve=True,
        )
        assert False
    except ArgumentTypeError:
        pass

    try:
        run(
            input_domains=["admin.hadrian.io"],
            temperature=-0.1,
            num_predictions=1,
            no_resolve=True,
        )
        assert False
    except ArgumentTypeError:
        pass



import subprocess


def test_():
    with open("README.md", "r") as f:
        readme = f.read()

    result = subprocess.run(
        ["python", "-m", "subwiz.cli", "-h"],
        capture_output=True,
        text=True,
        env={**dict(subprocess.os.environ), "COLUMNS": "100"},
    )
    help_text = result.stdout

    print(help_text)

    assert help_text in readme



import asyncio

import aiodns
import idna.core

from subwiz.type import Domain


NAME_SERVERS = ["1.1.1.1", "1.0.0.1", "8.8.8.8"]
TIMEOUT = 3
TRIES = 1


async def get_registered_domains(
    domains_to_check: set[Domain], resolution_concurrency: int
) -> set[Domain]:

    semaphore = asyncio.Semaphore(resolution_concurrency)
    resolver = aiodns.DNSResolver(
        nameservers=NAME_SERVERS, timeout=TIMEOUT, tries=TRIES
    )

    domains_list = list(domains_to_check)
    tasks = [dom.is_registered(resolver, semaphore) for dom in domains_to_check]
    results = await asyncio.gather(*tasks)

    return {dom for dom, is_reg in zip(domains_list, results) if is_reg}



"""
This code is largely copied from the nanogpt repository by Andrej Karpathy.
The generate method has been modified to output many sequences.

https://github.com/karpathy/nanoGPT

Full definition of a GPT Language Model, all of it in this single file.
References:
1) the official GPT-2 TensorFlow implementation released by OpenAI:
https://github.com/openai/gpt-2/blob/master/src/model.py
2) huggingface/transformers PyTorch implementation:
https://github.com/huggingface/transformers/blob/main/src/transformers/models/gpt2/modeling_gpt2.py
"""

import re
import math
import uuid

from pydantic import BaseModel, ConfigDict
import torch
import torch.nn as nn
from torch.nn import functional as F
from transformers import PreTrainedTokenizerFast
from typing import Callable, Optional


VALID_SUBDOMAIN_RE = re.compile(
    r"^(?!-)([a-z0-9](?:[a-z0-9-]{0,61}[a-z0-9])?)(?:\.([a-z0-9](?:[a-z0-9-]{0,61}[a-z0-9])?))*(?<!\.)$",
    re.IGNORECASE,
)

VALID_SUBDOMAIN_START_RE = re.compile(
    r"^(?:[A-Za-z0-9](?:[A-Za-z0-9-]{0,61}[A-Za-z0-9])?\.)*(?:[A-Za-z0-9](?:[A-Za-z0-9-]{0,61})?)?$",
    re.IGNORECASE,
)


class LayerNorm(nn.Module):
    """LayerNorm but with an optional bias. PyTorch doesn't support simply bias=False"""

    def __init__(self, ndim, bias):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(ndim))
        self.bias = nn.Parameter(torch.zeros(ndim)) if bias else None

    def forward(self, input):
        return F.layer_norm(input, self.weight.shape, self.weight, self.bias, 1e-5)


class CausalSelfAttention(nn.Module):
    def __init__(self, config):
        super().__init__()
        assert config.n_embd % config.n_head == 0
        # key, query, value projections for all heads, but in a batch
        self.c_attn = nn.Linear(config.n_embd, 3 * config.n_embd, bias=config.bias)
        # output projection
        self.c_proj = nn.Linear(config.n_embd, config.n_embd, bias=config.bias)
        # regularization
        self.attn_dropout = nn.Dropout(config.dropout)
        self.resid_dropout = nn.Dropout(config.dropout)
        self.n_head = config.n_head
        self.n_embd = config.n_embd
        self.dropout = config.dropout
        # flash attention make GPU go brrrrr but support is only in PyTorch >= 2.0
        self.flash = hasattr(torch.nn.functional, "scaled_dot_product_attention")
        if not self.flash:
            print(
                "WARNING: using slow attention. Flash Attention requires PyTorch >= 2.0"
            )
            # causal mask to ensure that attention is only applied to the left in the input sequence
            self.register_buffer(
                "bias",
                torch.tril(torch.ones(config.block_size, config.block_size)).view(
                    1, 1, config.block_size, config.block_size
                ),
            )

    def forward(self, x):
        (
            B,
            T,
            C,
        ) = x.size()  # batch size, sequence length, embedding dimensionality (n_embd)

        # calculate query, key, values for all heads in batch and move head forward to be the batch dim
        q, k, v = self.c_attn(x).split(self.n_embd, dim=2)
        k = k.view(B, T, self.n_head, C // self.n_head).transpose(
            1, 2
        )  # (B, nh, T, hs)
        q = q.view(B, T, self.n_head, C // self.n_head).transpose(
            1, 2
        )  # (B, nh, T, hs)
        v = v.view(B, T, self.n_head, C // self.n_head).transpose(
            1, 2
        )  # (B, nh, T, hs)

        # causal self-attention; Self-attend: (B, nh, T, hs) x (B, nh, hs, T) -> (B, nh, T, T)
        if self.flash:
            # efficient attention using Flash Attention CUDA kernels
            y = torch.nn.functional.scaled_dot_product_attention(
                q,
                k,
                v,
                attn_mask=None,
                dropout_p=self.dropout if self.training else 0,
                is_causal=True,
            )
        else:
            # manual implementation of attention
            att = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(k.size(-1)))
            att = att.masked_fill(self.bias[:, :, :T, :T] == 0, float("-inf"))
            att = F.softmax(att, dim=-1)
            att = self.attn_dropout(att)
            y = att @ v  # (B, nh, T, T) x (B, nh, T, hs) -> (B, nh, T, hs)
        y = (
            y.transpose(1, 2).contiguous().view(B, T, C)
        )  # re-assemble all head outputs side by side

        # output projection
        y = self.resid_dropout(self.c_proj(y))
        return y


class MLP(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.c_fc = nn.Linear(config.n_embd, 4 * config.n_embd, bias=config.bias)
        self.gelu = nn.GELU()
        self.c_proj = nn.Linear(4 * config.n_embd, config.n_embd, bias=config.bias)
        self.dropout = nn.Dropout(config.dropout)

    def forward(self, x):
        x = self.c_fc(x)
        x = self.gelu(x)
        x = self.c_proj(x)
        x = self.dropout(x)
        return x


class Block(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.ln_1 = LayerNorm(config.n_embd, bias=config.bias)
        self.attn = CausalSelfAttention(config)
        self.ln_2 = LayerNorm(config.n_embd, bias=config.bias)
        self.mlp = MLP(config)

    def forward(self, x):
        x = x + self.attn(self.ln_1(x))
        x = x + self.mlp(self.ln_2(x))
        return x


class GPTConfig(BaseModel):
    block_size: int = 1024
    vocab_size: int = (
        50304  # GPT-2 vocab_size of 50257, padded up to nearest multiple of 64 for efficiency
    )
    n_layer: int = 12
    n_head: int = 12
    n_embd: int = 768
    dropout: float = 0.0
    bias: bool = (
        True  # True: bias in Linears and LayerNorms, like GPT-2. False: a bit better and faster
    )
    tokenizer_file: str = "resources/tokenizer.json"

    model_config = ConfigDict(extra="ignore")


class GPT(nn.Module):
    def __init__(self, config: GPTConfig):
        super().__init__()
        assert config.vocab_size is not None
        assert config.block_size is not None
        self.config = config
        self.tokenizer = PreTrainedTokenizerFast(
            tokenizer_file=config.tokenizer_file, clean_up_tokenization_spaces=True
        )
        self.end_token = self.tokenizer("[END]")["input_ids"][0]
        self.comma_token = self.tokenizer(",")["input_ids"][0]
        self.delim_token = self.tokenizer("[DELIM]")["input_ids"][0]

        self.transformer = nn.ModuleDict(
            dict(
                wte=nn.Embedding(config.vocab_size, config.n_embd),
                wpe=nn.Embedding(config.block_size, config.n_embd),
                drop=nn.Dropout(config.dropout),
                h=nn.ModuleList([Block(config) for _ in range(config.n_layer)]),
                ln_f=LayerNorm(config.n_embd, bias=config.bias),
            )
        )
        self.lm_head = nn.Linear(config.n_embd, config.vocab_size, bias=False)
        # with weight tying when using torch.compile() some warnings get generated:
        # "UserWarning: functional_call was passed multiple values for tied weights.
        # This behavior is deprecated and will be an error in future versions"
        # not 100% sure what this is, so far seems to be harmless. TODO investigate
        self.transformer.wte.weight = (
            self.lm_head.weight
        )  # https://paperswithcode.com/method/weight-tying

        # init all weights
        self.apply(self._init_weights)
        # apply special scaled init to the residual projections, per GPT-2 paper
        for pn, p in self.named_parameters():
            if pn.endswith("c_proj.weight"):
                torch.nn.init.normal_(
                    p, mean=0.0, std=0.02 / math.sqrt(2 * config.n_layer)
                )

        # report number of parameters
        # print("number of parameters: %.2fM" % (self.get_num_params() / 1e6,))

    def get_num_params(self, non_embedding=True):
        """
        Return the number of parameters in the model.
        For non-embedding count (default), the position embeddings get subtracted.
        The token embeddings would too, except due to the parameter sharing these
        params are actually used as weights in the final layer, so we include them.
        """
        n_params = sum(p.numel() for p in self.parameters())
        if non_embedding:
            n_params -= self.transformer.wpe.weight.numel()
        return n_params

    def _init_weights(self, module):
        if isinstance(module, nn.Linear):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
            if module.bias is not None:
                torch.nn.init.zeros_(module.bias)
        elif isinstance(module, nn.Embedding):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)

    def forward(self, idx, targets=None):
        # with torch.autograd.detect_anomaly():
        #     if torch.isnan(idx).any():
        #         print(f'NAN found!: {idx}')

        device = idx.device
        b, t = idx.size()
        assert (
            t <= self.config.block_size
        ), f"Cannot forward sequence of length {t}, block size is only {self.config.block_size}"
        pos = torch.arange(0, t, dtype=torch.long, device=device)  # shape (t)

        # forward the GPT model itself
        tok_emb = self.transformer.wte(idx)  # token embeddings of shape (b, t, n_embd)
        pos_emb = self.transformer.wpe(pos)  # position embeddings of shape (t, n_embd)
        x = self.transformer.drop(tok_emb + pos_emb)
        for block in self.transformer.h:
            x = block(x)
        x = self.transformer.ln_f(x)

        if targets is not None:
            # if we are given some desired targets also calculate the loss
            logits = self.lm_head(x)
            loss = F.cross_entropy(
                logits.view(-1, logits.size(-1)), targets.view(-1), ignore_index=-1
            )
        else:
            # inference-time mini-optimization: only forward the lm_head on the very last position
            logits = self.lm_head(
                x[:, [-1], :]
            )  # note: using list [-1] to preserve the time dim
            loss = None

        return logits, loss

    def crop_block_size(self, block_size):
        # model surgery to decrease the block size if necessary
        # e.g. we may load the GPT2 pretrained model checkpoint (block size 1024)
        # but want to use a smaller block size for some smaller, simpler model
        assert block_size <= self.config.block_size
        self.config.block_size = block_size
        self.transformer.wpe.weight = nn.Parameter(
            self.transformer.wpe.weight[:block_size]
        )
        for block in self.transformer.h:
            if hasattr(block.attn, "bias"):
                block.attn.bias = block.attn.bias[:, :, :block_size, :block_size]

    @property
    def device(self) -> str:
        # assign model inputs to the right device
        return next(self.lm_head.parameters()).device.type

    @torch.no_grad()
    def generate(
        self,
        idx: torch.Tensor,
        max_new_tokens: int = 12,
        temperature: float = 0.0,
        topn: int = 100,
        pruning_ratio: float = 4,
        pruning_offset: float = 5,
        log_file: Optional[str] = None,
        on_iteration: Callable = None,
        blocked_outputs: set[str] = None,  # doesn't output these strings
    ) -> torch.Tensor:
        """Custom generate function that outputs topn sequences, different to the original nanoGPT implementation."""

        if topn <= 0:
            raise ValueError("topn should be greater than 0")

        if not 0 < max_new_tokens <= 20:
            raise ValueError("max_new_tokens should be in (0, 20]")

        run_uuid = uuid.uuid4()

        idx = idx.to(self.device)
        sequences = idx.unsqueeze(0)

        probabilities = torch.tensor([1.0], device=self.device)

        finished_sequences = torch.tensor([], device=self.device)
        finished_probs = torch.tensor([], device=self.device)

        # compute number of sequences to pass to each iteration
        sequences_per_iter = round(pruning_offset + topn / pruning_ratio)

        for i in range(max_new_tokens):
            if on_iteration is not None:
                on_iteration()

            # trim the sequences down to block size
            sequences = sequences[:, -self.config.block_size :]

            # remove any invalid subdomain starts
            outputs = self.tokenizer.batch_decode(sequences[:, -i:])
            outputs = [o.replace(" ", "") for o in outputs]
            outputs = [o.rsplit("[DELIM]", 1)[-1] for o in outputs]
            valid_start_mask = [
                bool(VALID_SUBDOMAIN_START_RE.match(o)) for o in outputs
            ]
            sequences = sequences[valid_start_mask]
            probabilities = probabilities[valid_start_mask]
            outputs = [o for o, valid in zip(outputs, valid_start_mask) if valid]

            # inference the model in batches
            batch_size = 8
            logits, _ = self(sequences[:batch_size])
            for j in range(batch_size, len(sequences), batch_size):
                new_logits, _ = self(sequences[j : j + batch_size])
                logits = torch.cat(tensors=(new_logits, logits), dim=0)
            logits = logits.squeeze(1)

            # take N most probable next tokens for each sequence
            output_probs = F.softmax(logits, dim=-1)
            new_sequence_probs = output_probs * probabilities.unsqueeze(1)

            # remove finished sequences (after end token) and cache their probs
            if i > 0:
                # likelihood of each sequence having an end or comma token next
                comma_token_probs = new_sequence_probs[:, self.comma_token]
                end_token_probs = new_sequence_probs[:, self.end_token]
                _finish_probs = end_token_probs + comma_token_probs
                _finished_sequences = sequences.clone().detach()

                # don't output any invalid subdomains
                valid_subdomain_mask = [
                    bool(VALID_SUBDOMAIN_RE.match(o)) for o in outputs
                ]
                _finish_probs = _finish_probs[valid_subdomain_mask]
                _finished_sequences = _finished_sequences[valid_subdomain_mask]
                outputs = [o for o, val in zip(outputs, valid_subdomain_mask) if val]

                # don't output anything in blocked_outputs
                if blocked_outputs and outputs:
                    blocked_outputs_mask = torch.tensor(
                        [s not in blocked_outputs for s in outputs], device=self.device
                    )
                    _finish_probs = _finish_probs[blocked_outputs_mask]
                    _finished_sequences = _finished_sequences[blocked_outputs_mask]

                # add all sequences to the finished_sequences with prob of ending
                finished_sequences = torch.cat(
                    (finished_sequences, _finished_sequences)
                )
                finished_probs = torch.cat((finished_probs, _finish_probs), dim=-1)

            # remove sequences and tokens with a probability that is too low
            if len(finished_sequences) > topn:
                # torch.kthvalue is not implemented on MPS, so we use topk
                lowest_viable_probability = torch.topk(finished_probs, topn).values[-1]
                viable_sequences = probabilities >= lowest_viable_probability

                if viable_sequences.sum() == 0:
                    break

                # remove sequences with a too low probability
                sequences = sequences[viable_sequences]
                probabilities = probabilities[viable_sequences]
                logits = logits[viable_sequences]
                new_sequence_probs = new_sequence_probs[viable_sequences]

                # remove tokens that would generate sequences with too low probability
                token_mask = new_sequence_probs < lowest_viable_probability
                if token_mask.sum() == 0:
                    break

                new_sequence_probs[token_mask] = 0
                logits[token_mask] = 0

            # do not sample the end token or comma token for the next iter
            new_sequence_probs[:, self.end_token] = 0
            new_sequence_probs[:, self.comma_token] = 0

            # number of sequences to pass to next iteration
            num_nonzero_probs = torch.count_nonzero(new_sequence_probs).item()
            num_seqs_next_iter = min(sequences_per_iter, num_nonzero_probs)

            if num_seqs_next_iter == 0:
                break

            if temperature == 0:  # select most likely tokens for next iteration
                new_sequence_probs = new_sequence_probs.flatten()
                _, idx_next = torch.topk(new_sequence_probs, num_seqs_next_iter)

            else:  # sample tokens for next iteration
                # recalculate probabilities using temperature
                scaled_logits = logits / (temperature + 1e-1)
                probs_with_temp = F.softmax(scaled_logits, dim=-1)
                probs_with_temp = probs_with_temp * probabilities.unsqueeze(1)

                probs_with_temp[:, self.end_token] = 0
                probs_with_temp[:, self.comma_token] = 0

                # sample tokens for next iteration
                probs_with_temp = probs_with_temp.flatten()
                probs_with_temp[probs_with_temp < 0] = 0
                idx_next = torch.multinomial(probs_with_temp, num_seqs_next_iter)

            # add the sampled tokens to the end of each sequence
            sequence_idx = idx_next // self.config.vocab_size
            token_values = idx_next % self.config.vocab_size

            sequences = sequences[sequence_idx]
            sequences = torch.cat([sequences, token_values.unsqueeze(1)], dim=-1)
            probabilities = new_sequence_probs.flatten()[idx_next]

            if log_file is not None:
                _, current_best_idx = torch.topk(
                    finished_probs, min(topn, len(finished_probs))
                )
                current_best = finished_sequences[current_best_idx]
                self.log_generation_data(
                    log_file=log_file,
                    run_id=run_uuid,
                    topn=topn,
                    x=idx,
                    iteration=i,
                    probabilities=probabilities,
                    current_preds=current_best,
                    finished_probs=finished_probs,
                )

        if len(finished_sequences) <= topn:
            return finished_sequences

        _, final_indices = torch.topk(finished_probs, topn)
        final_sequences = finished_sequences[final_indices]

        return final_sequences

    @staticmethod
    def from_checkpoint(
        path: str,
        return_train_params: bool = False,
        device: str = "cpu",
        tokenizer_path: Optional[str] = None,
    ):
        checkpoint = torch.load(path, map_location=device, weights_only=True)

        config = GPTConfig(**checkpoint["model_args"])
        if tokenizer_path:
            config.tokenizer_file = tokenizer_path
        model = GPT(config)
        state_dict = checkpoint["model"]

        # fix the keys of the state dictionary :(
        # honestly no idea how checkpoints sometimes get this prefix, have to debug more
        unwanted_prefix = "_orig_mod."
        for k, v in list(state_dict.items()):
            if k.startswith(unwanted_prefix):
                state_dict[k[len(unwanted_prefix) :]] = state_dict.pop(k)
        model.load_state_dict(state_dict)
        model.to(device)

        if not return_train_params:
            return model

        iter_num = checkpoint["iter_num"]
        best_val_loss = checkpoint["best_val_loss"]
        optim_state = checkpoint["optimizer"]

        assert isinstance(iter_num, int)
        assert isinstance(best_val_loss, torch.Tensor)
        assert isinstance(optim_state, dict)

        return model, iter_num, best_val_loss, optim_state



import asyncio

from subwiz.resolve import get_registered_domains
from subwiz.type import Domain


def test_():
    domain_strings = {"api.hadrian.io", "app.hadrian.io", "random_string.hadrian.io"}
    input_domains = {Domain(dom) for dom in domain_strings}
    registered_domains = asyncio.run(
        get_registered_domains(input_domains, resolution_concurrency=10)
    )
    assert registered_domains == {Domain("api.hadrian.io"), Domain("app.hadrian.io")}



import argparse
import asyncio
import os
import re
from collections import defaultdict
from typing import Callable

from huggingface_hub import hf_hub_download
from huggingface_hub.utils import disable_progress_bars, enable_progress_bars
import torch
from transformers import PreTrainedTokenizerFast

from subwiz.cli_printer import print_hello, print_log, print_progress_dot
from subwiz.model import GPT
from subwiz.resolve import get_registered_domains
from subwiz.type import (
    Domain,
    input_domains_type,
    device_type,
    max_recursion_type,
    positive_int_type,
    temperature_type,
    concurrency_type,
)


MODEL_REPO = "HadrianSecurity/subwiz"
MODEL_FILE = "model.pt"
TOKENIZER_FILE = "tokenizer.json"
CONFIG_FILE = "config.json"


def get_model_and_tokenizer(
    force_download: bool,
    device: str,
    quiet: bool,
) -> tuple[GPT, PreTrainedTokenizerFast]:
    """Download files from HuggingFace to run subwiz. Caches in local file system."""
    if quiet:
        disable_progress_bars()

    model_path = hf_hub_download(
        repo_id=MODEL_REPO, filename=MODEL_FILE, force_download=force_download
    )
    tokenizer_path = hf_hub_download(
        repo_id=MODEL_REPO, filename=TOKENIZER_FILE, force_download=force_download
    )
    hf_hub_download(
        repo_id=MODEL_REPO, filename=CONFIG_FILE, force_download=force_download
    )
    if quiet:
        enable_progress_bars()

    gpt_model = GPT.from_checkpoint(
        model_path, device=device, tokenizer_path=tokenizer_path
    )
    tokenizer = PreTrainedTokenizerFast(
        tokenizer_file=tokenizer_path, clean_up_tokenization_spaces=True
    )

    return gpt_model, tokenizer


def run_inference(
    input_domains: set[Domain],
    gpt_model: GPT,
    tokenizer: PreTrainedTokenizerFast,
    num_predictions: int,
    max_new_tokens: int,
    temperature: float,
    blocked_domains: set[Domain],
    on_inference_iteration: Callable = None,
) -> set[Domain]:
    """Preprocess inputs, tokenize text, run inference and decode tokens back to text."""

    apex = next(iter(input_domains)).apex_domain
    subs = [dom.subdomain for dom in input_domains]
    tokenizer_input = ",".join(sorted(subs)) + "[DELIM]"
    # TODO: pick a different subset, if some were out of context last iteration

    x = tokenizer.encode(tokenizer_input)
    x = [1] * (gpt_model.config.block_size - len(x)) + x
    x = torch.tensor(x)

    blocked_outputs = {dom.subdomain for dom in blocked_domains}

    predictions = gpt_model.generate(
        x,
        max_new_tokens=max_new_tokens,
        topn=num_predictions,
        on_iteration=on_inference_iteration,
        temperature=temperature,
        blocked_outputs=blocked_outputs,
    )
    predictions = predictions.int().tolist()

    predictions = {
        tokenizer.decode(pred).replace(" ", "").rsplit("[DELIM]", 1)[1]
        for pred in predictions
    }

    predictions: set[str] = {sub + "." + apex for sub in predictions}

    predicted_domains: set[Domain] = set()
    for pred in predictions:
        try:
            predicted_domains.add(Domain(pred))
        except ValueError:  # invalid domain name
            pass

    return predicted_domains


def _get_domains_for_group(
    domains_in_group: set[Domain],
    all_apexes: set[str],
    gpt_model: GPT,
    tokenizer: PreTrainedTokenizerFast,
    multi_apex: bool,
    num_predictions: int,
    max_new_tokens: int,
    max_recursion: int,
    temperature: float,
    no_resolve: bool,
    resolution_concurrency: int,
    quiet: bool,
) -> set[str]:
    """For a group of subdomains that share an apex: run inference and check if they resolve, recursively."""

    blocked_domains: set[Domain] = domains_in_group.copy()
    all_predictions_that_resolve: set[Domain] = set()
    apex = next(iter(domains_in_group)).apex_domain

    for i in range(max_recursion):

        on_inference_iteration = None
        if not quiet:

            dots_emitted = 0

            def _counting_progress_dot():
                nonlocal dots_emitted
                dots_emitted += 1
                print_progress_dot()

            on_inference_iteration = _counting_progress_dot

            log = "running inference"
            if multi_apex:
                log += f" for {apex}"
            if i:
                log += f" x{i + 1}"
            print_log(log, end="")

        predictions = run_inference(
            input_domains=domains_in_group,
            gpt_model=gpt_model,
            tokenizer=tokenizer,
            num_predictions=num_predictions,
            max_new_tokens=max_new_tokens,
            temperature=temperature,
            on_inference_iteration=on_inference_iteration,
            blocked_domains=blocked_domains,
        )

        if no_resolve:
            if not quiet:
                print_log("", end="\n")
            return {str(dom) for dom in predictions}

        predictions_that_resolve = asyncio.run(
            get_registered_domains(predictions, resolution_concurrency)
        )

        if not quiet:
            if not max_recursion:
                end_log = ""
            else:
                printed_characters = len(apex) + dots_emitted + (0 if i == 0 else 3)
                max_characters = max(len(a) for a in all_apexes) + max_new_tokens + 3
                extra_spaces = max_characters - printed_characters + 1

                sub_count = len(predictions_that_resolve)
                subs_label = "subdomain" if sub_count == 1 else "subdomains"
                end_log = " " * extra_spaces + f"found {sub_count:>3} {subs_label:>10}"
            print_log(end_log, end="\n")

        if not predictions_that_resolve:
            break

        all_predictions_that_resolve |= predictions_that_resolve

        blocked_domains |= predictions
        domains_in_group |= predictions_that_resolve

    return {str(dom) for dom in all_predictions_that_resolve}


def run(
    input_domains: list[str],
    device: str = "auto",
    num_predictions: int = 500,
    max_new_tokens: int = 10,
    temperature: float = 0,
    resolution_concurrency: int = 128,
    no_resolve: bool = False,
    force_download: bool = False,
    multi_apex: bool = False,
    max_recursion: int = 5,
    quiet: bool = True,
) -> list[str]:
    """Check types, download model, get new subdomains for each apex."""
    if not quiet:
        print_hello()

    domain_objects = input_domains_type(input_domains)
    device = device_type(device)
    num_predictions = positive_int_type(num_predictions)
    max_new_tokens = positive_int_type(max_new_tokens)
    max_recursion = max_recursion_type(max_recursion)
    temperature = temperature_type(temperature)
    resolution_concurrency = concurrency_type(resolution_concurrency)

    domain_groups = defaultdict(set)
    for dom in domain_objects:
        domain_groups[dom.apex_domain].add(dom)

    if len(domain_groups) > 1 and not multi_apex:
        raise argparse.ArgumentTypeError(
            f"multiple apex domains found: {sorted(domain_groups.keys())}. "
            "Use the --multi-apex flag to process them all."
        )

    gpt_model, tokenizer = get_model_and_tokenizer(
        force_download, device=device, quiet=quiet
    )
    found_domains = set()

    for apex in sorted(domain_groups):
        found_domains |= _get_domains_for_group(
            domains_in_group=domain_groups[apex],
            all_apexes=set(domain_groups.keys()),
            gpt_model=gpt_model,
            tokenizer=tokenizer,
            multi_apex=multi_apex,
            num_predictions=num_predictions,
            max_new_tokens=max_new_tokens,
            max_recursion=max_recursion,
            temperature=temperature,
            no_resolve=no_resolve,
            resolution_concurrency=resolution_concurrency,
            quiet=quiet,
        )

    return sorted(found_domains)
